package cs155.hw3;


import java.awt.*;
import javax.swing.*;


/** 
  This consists of static methods for drawing a scene on a canvas.
  It also contains all the auxilliary methods for intersecting
  rays with objects, scenes, etc.
**/

public class RayTracer3D {
 
    public static int recursionDepth=4;


    // this insersects the specified ray with all objects in the scene
    /**
       Return the HitRay object for the closest intersection of the ray with an object in the scene
     **/
    public static RayHit intersectScene(Scene3D s, Ray3D r) {
	return s.objs.rayIntersect(r);
    }


    /**
       Compute the color corresponding of the pixel corresponding to the RayHit intersection "hit"
    **/
    public static Color3D computeColor(Ray3D r, Scene3D s, int depth) {

	RayHit hit = intersectScene(s,r);


        if (hit.distance == Double.POSITIVE_INFINITY)
	    return s.backgroundColor;


	Point3D n = hit.normal;
	Object3D obj = hit.obj;

        Point3D p = r.atTime(hit.distance);  // intersection point of ray with an object in the scene


	boolean outside = (n.dot(r.d) < 0);
	if (!outside) n = n.scalarMult(-1d);

	Material m = (outside?obj.outsideMat:obj.insideMat);
            
	Color3D pixelColor = m.emissive.add(s.ambient.times(m.ambient));


	for (int k = 0; k<s.numLights; k++) {
		  /* here we calculate the color contribution for the kth light
		     but this is pretty ugly code! I wonder if we can clean it up
		     by moving some more functionality to the Light3D class
		  */

                  Color3D localColor = Color3D.BLACK;
                  Light3D light = s.light[k];
                  Point3D lightPos = light.location;
                  Point3D lightVec = lightPos.subtract(p);
                  double lightDist = lightVec.length(); // distance from point to the light
                  lightVec = lightVec.normalize(); //vector from point to the light
                  Ray3D LR = new Ray3D(p,lightVec);

                  // check to see if the light casts a shadow over this point
                  RayHit lightHit = intersectScene(s,LR);
                  if (lightHit.distance < lightDist-Object3D.epsilon) 
		      continue; // skip the rest of the body of this loop


                  Point3D eyeVec = r.p.subtract(p).normalize();

                  Color3D localAmbient  = m.ambient.times(light.ambient);

                  double diffuseIntensity = light.diffuse(lightVec,n); 
                  Color3D localDiffuse  = m.diffuse.times(light.diffuse).scale(diffuseIntensity);

                  double specularIntensity = light.specular(lightVec,n,eyeVec,m.hardness); 
                  Color3D localSpecular  = m.specular.times(light.specular).scale(specularIntensity);

                  localColor = localAmbient.add(localDiffuse).add(localSpecular);
                  localColor = localColor.scale(light.intensity).scale(1-m.reflect-m.refract).scale(light.attenuation(lightDist));

                  // calculate reflected color
                  Ray3D reflectedRay = reflect(p,n,r);
                  Color3D reflectedColor = Color3D.BLACK;
		  if (depth > 0 )
                     reflectedColor= computeColor(reflectedRay, s, depth-1).scale(m.reflect);

		  // check to see if the light is blocked 
		  pixelColor = pixelColor.add(localColor).add(reflectedColor);
	      }

	return(pixelColor);
    }



    public static Ray3D reflect(Point3D p, Point3D N, Ray3D r) {
	return new Ray3D(p,r.d.subtract(N.scalarMult(2*N.dot(r.d))));
    }



    public static Point3D eye = new Point3D(0d,0d,0d);

    /**
     draw a  view of the scene on a canvas using the drawPixel command.
     flush the image from the offscreen buffer to the canvas using the refresh command.
     The camera screen is in the plane z=-1 in the interval [-1,1]x[-1,1]
     The eye is initially at (0,0,0) looking toward the screen
    **/
  public static void drawScene(Scene3D s,Canvas3D canvas) {
        double 
          h = canvas.height(),
          w = canvas.width(); 

        for(int i=0;i<w; i++) { 
      	    if ((i % 100)==0) // this helps us see how fast the program is computing pixels
                {System.out.print(" "+i);System.out.flush(); canvas.refresh();}
            
          for(int j=0; j<h; j++)  {

            // calculate the direction of the ray, d=(u,v,-1), in world coordinates 
            // that corresponds to the pixel at position (i,j) in screen coordinates
            double u,v; 
            u = (double)((i-w/2)/(w/2));
            v = (double)((h/2-j)/(w/2)); // the width is 2 but height varies depending on screen size ...

            // calculate the color of the pixel corresponding to that ray
            Color3D pixelColor = Color3D.BLACK; 
            for (int k=0; k< s.OSI; k++) {
		Ray3D r1 = s.camera.generateRay(u+(Math.random()-0.5)/(w/2), v+(Math.random()-0.5)/(h/2));
		Color3D localColor = computeColor(r1,s,recursionDepth); 
		pixelColor = pixelColor.add(localColor);
            }
            pixelColor = pixelColor.scale(1.0/s.OSI);
	    canvas.drawPixel(i,j,pixelColor.toColor());
         } // close for i
       } // close for j
     } // close drawScene
}



   

